/*----------------------------------------------------------------*
 * SUGI PAPER 326-2010
 * Adding Statistical Functionality to the DATA Step with FCMP
 *
 * BY: Stacey Christian and Jacques Rioux
 *
 * EXAMPLE 2: Iterated Reweighted Least Squares Algorithm 
 *---------------------------------------------------------------*/

/*---------------------------------------------------------------*
 * Macro to prepare data for the Iterated Weighted Least Squares 
 * Algorithm. Creates indicator variables for each time 
 * interval and calculates 
 *       log_log_pij = log(-log(prob of survival))
 *
 * Parameters:                                                    
 *  data          : the dataset on which to run the PH model      
 *  count         : name of the count variable                    
 *  time          : name of the time variable                     
 *  regressors    : list of regression variables                  
 *  outData       : output dataset to be created                  
 *  intervalCount : number of time intervals (OUT)                
 *  intervalVars  : Interval_1-Interval_N (OUT)                  
 *---------------------------------------------------------------*/
%macro prepare_phdata_macro;
   %let data         = %sysfunc( dequote(&data ));
   %let count        = %sysfunc( dequote(&count ));
   %let time         = %sysfunc( dequote(&time ));
   %let regressors   = %sysfunc( dequote(&regressors ));
   %let outData      = %sysfunc( dequote(&outData ));
   %let intervalVars = %sysfunc( dequote(&intervalVars ));

   proc sort data=&data out =&outData;
      by &regressors &time;
   run;

   proc sql noprint;
      select distinct &time 
      into :times separated by " "
      from &outData
      order by &time
      ;
      select count(distinct &time) 
      into :timeCount
      from &outData
      order by &time
      ;
      select max(&time) 
      into :MaxTime
      from &outData
      ;
   quit;

   %let IntervalCount =  %eval(&timeCount-1);

   data &outData;
      set &outData;
      retain LastCount 0 LastTime &MaxTime;

      array TimeIndicator[&IntervalCount] Interval_1-Interval_%trim(&IntervalCount);
      array Times[&TimeCount] _TEMPORARY_ (&Times);

      if &Time > LastTime then do;
         do i = 1 to &IntervalCount;
            if &Time = Times[i+1] then do;
               TimeIndicator[i] = 1;
            end;
            else do;
               TimeIndicator[i] = 0;
            end;
         end;

         Nij         = LastCount;
         Pij         = &Count/LastCount;
         Log_Log_Pij = log(-log(&Count/LastCount));
         Weight      = 1.0;
         output;
      end;

      LastCount = &Count;
      LastTime  = &Time;

      drop  &Count i LastTime LastCount;
   run;

   %let intervalVars = Interval_1-Interval_%trim(&IntervalCount);

%mend prepare_phdata_macro;


/*---------------------------------------------------------------*
 * Macro that uses proc REG to run regression                    
 *                                                               
 * Parameters:                                                   
 *  data        : the dataset on which to run the regression      
 *  params      : output dataset for the fitted parameters       
 *  dependent   : name of the dependent variable                  
 *  independent : space delimited list of independent variables   
 *  weight      : name of the weighting variable                 
 *---------------------------------------------------------------*/
%macro run_regression_macro;
   %let data        = %sysfunc( dequote(&data)        );
   %let params      = %sysfunc( dequote(&params)      );
   %let dependent   = %sysfunc( dequote(&dependent)   );
   %let independent = %sysfunc( dequote(&independent) );
   %let weight      = %sysfunc( dequote(&weight)      );

   proc reg data=&data outest=&params NOPRINT;
      model  &dependent = &independent/noint;
      weight &weight;
   quit;

   data &params;
      set &params;
      keep &independent;
   run;
%mend run_regression_macro;


/*-----------------------------------------------------------------*
 * Macro to recompute the regression weight at every      
 * iteration in data set passed.  
 *                                                                
 * Parameters:                                                    
 *  data        : the dataset on which to reweight      
 *  params      : output parameter data set from proc reg                    
 *  regressors  : name of the regression variables                      
 *  weight      : name of the weighting variable                 
 *----------------------------------------------------------------*/
%macro update_weights_macro;
   %let data       = %sysfunc( dequote(&data)      );
   %let params     = %sysfunc( dequote(&params)    );
   %let regressors = %sysfunc( dequote(&regressors)); 
   %let weight     = %sysfunc( dequote(&weight)); 

   data &params._Params; set &params;
      /* set Param_x1 = x1; Param_x2 = x2; ... */
      %let nVar = 1;
      %let Var = %scan(&regressors, &nVar, ' ');
      %do %while ("&Var" ne "");
          Param_&Var.    = &Var;   
          %let nVar = %eval(&nVar + 1);
          %let Var = %scan(&regressors, &nVar, ' ');
      %end;
      /* set Param_Interval_1 = Interval_1; ... */
      %do i = 1 %to &intervalCount;
         Param_Interval_&i = Interval_&i;
      %end;

      keep Param_:;
   run;
   /* merge params with data */
   proc sql noprint;
      create table _temp_ as 
      select * from &data, &params._Params;
   quit;
   /* re-calculate weight  */
   data &data; set _temp_;
      retain LastSj  1 lastTime 0;
      array Param_Interval[&intervalCount] Param_Interval_1 -
                                           Param_Interval_&intervalCount;
      array Interval[&intervalCount] Interval_1 - Interval_&intervalCount;

      Pj = 0;
      do i = 1 to &intervalCount.;
        Pj = Pj  + exp(-exp(Param_Interval[i])) * Interval[i];
      end;

      if _N_ = 1 then do;
            Sj = Pj;
      end;
      else do;
         if &time < lastTime then do;
             Sj = Pj ;
         end;
         else do;
             Sj = LastSj * Pj;
         end;
      end;

      /* ExpXB = exp(Param_x1 * x1 + Param_x2 * x2 + ... + Param_xn * xn); */
      ExpXB  = exp(0
                %let nVar = 1;
                %let Var = %scan(&regressors, &nVar, ' ');
                %do %while ("&Var" ne "");
                    + Param_&Var. * &Var   
                    %let nVar = %eval(&nVar + 1);
                    %let Var = %scan(&regressors, &nVar, ' ');
                %end;
          );

      &Weight = ( 1 - Pj**ExpXB ) / Nij / Pj**ExpXB / ExpXB**2 / log(Pj)**2;  

      LastSj = Sj;
      LastTime = &time;

      drop LastTime LastSj i Param_:;
   run;

%mend update_weights_macro;

/*---------------------------------------------------------------*
 * FCMP Functions Wrappers for macros above.
 *---------------------------------------------------------------*/
proc fcmp outlib=sasuser.example2.mindist;

    subroutine prepare_phdata(data $, count $, time $, regressors $,
                              outData $, intervalCount, intervalVars $);
      outargs intervalCount, intervalVars;

      intervalVars = "";
      intervalCount = .;

      rc = run_macro ( 'prepare_phdata_macro', 
                       data          ,
                       count         , 
                       time          , 
                       regressors    ,  
                       outData       ,
                       intervalCount ,
                       intervalVars 
                       );
   endsub;

   subroutine run_regression( data $, dependent $, independent $, weight $, 
                              params $, paramArray[*]);
      outargs paramArray;

      array tmpArray[1] _temporary_;
      call dynamic_array(tmpArray,dim(paramArray));

      rc = RUN_MACRO ('run_regression_macro', data, params, dependent, independent, weight) ;
      rc = read_array(params, tmpArray);

      do i = 1 to dim(paramArray);
         paramArray[i] = tmpArray[1,i];
      end;
   endsub;

   subroutine update_weights( data $, params $, regressors $, weight $, intervalCount);
      rc = RUN_MACRO ('update_weights_macro', 
                      data, params, regressors, weight, intervalCount);
   endsub;

   function calc_max_relative_diff(params1[*],params2[*]);
         outargs params1;
         maxRelativeDifference = 0;
		 numParams = DIM(params1);
         do i = 1 to numParams;
            maxRelativeDifference = max( maxRelativeDifference, abs((params1[i]-params2[i])/params2[i]));
            params1[i] = params2[i];
         end;
         return(maxRelativeDifference);
   endsub;
run;

options cmplib=sasuser.example2;

/******************************************************************/
/*                                                                */
/* The main algorithm to implement the Iterated Reweighted Least  */
/* Square Estimator.                                              */
/*                                                                */
/* Parameters:                                                    */
/*  data        : the dataset on which to run the PH model        */
/*  outData     : output dataset for the fitted parameters        */
/*  count       : name of the count variable                      */
/*  time        : name of the time variable                       */
/*  regressors  : name of the weighting variable                  */
/*  paramds     : fitted parameters dataset name                  */
/*                                                                */
/******************************************************************/
%let numParams    = 15;
%let data         = "work.mdphdata";
%let preparedData = "work.outdata";
%let paramData    = "work.phresults";
%let regressors   = "x1 x2 x3";
%let regCount     = 3;
%let depVar       = "log_log_Pij";
%let weightVar    = "Weight";

data _NULL_;
      /*- initial data preparation -*/
      length intervalVars $ 128;
      intervalCount = 0; intervalVars = "";
      call prepare_phdata( &data, "Count", "Quarter", &regressors, &preparedData, 
                           intervalCount, intervalVars );
      independentVars = &regressors || " " || intervalVars;

      array params1[&numParams] _temporary_;
      array params2[&numParams] _temporary_;

      /*- initial regression run with initial weights -*/
      call run_regression( &preparedData, &depVar,independentVars, 
                           &weightVar, &paramData, params1 );

      /*- iterative loop -*/
      maxRelativeDifference = 1;
      do while( maxRelativeDifference > 0.0001 ); 
         call update_weights(&preparedData, &paramData, &regressors, &weightVar, intervalCount);
            
         call run_regression( &preparedData, &depVar,independentVars,
                              &weightVar, &paramData, params2 );  
 
         maxRelativeDifference = calc_max_relative_diff(params1,params2);
         put maxRelativeDifference=;
      end;

      call update_weights(&preparedData, &paramData, &regressors, &weightVar, intervalCount);
run;



