/*--------------------------------------------------------------*
 * SUGI PAPER 326-2010
 * Adding Statistical Functionality to the DATA Step with FCMP
 *
 * BY: Stacey Christian and Jacques Rioux
 *
 * EXAMPLE 1: Recursive Algorithm for Segmenting 
 *            Time Series Data
 *-------------------------------------------------------------*/

/*---------------------------------------------------*
 * Macro to get number of observations in a dataset
 *
 * Parameters:
 *    ds   - name of dataset get nobs for
 *    nobs - output
 *---------------------------------------------------*/
%macro get_nobs_macro;
   %let ds = %sysfunc(dequote(&ds));
   %let nobs=0;
   %let dsid=%sysfunc(open(&ds,I));
   %if &dsid ^= 0 %then %do;
      %let nobs=%sysfunc(attrn(&dsid,nlobs));
      %let rc=%sysfunc(close(&dsid));
   %end;
%mend get_nobs_macro;


/*---------------------------------------------------*
 * Macro to append segment information to dataset
 *
 * Parameters:
 *    seg_ds    - name of dataset append to
 *    first_obs - first row in segment
 *    last_obs  - last row in segment
 *    error     - approximation error for segment  
 *---------------------------------------------------*/
%macro append_segment_macro;
     %let seg_ds = %sysfunc(dequote(&seg_ds));
     data _temp_; 
       start = &first_obs;
       end   = &last_obs;
       error = &error;
     run;
     proc append base=&seg_ds data=_temp_ force;
     run;
%mend append_segment_macro;

/*---------------------------------------------------*
 * Macro to perform a linear approximation of a 
 * segment of time series data using PROC REG. 
 *
 * Parameters:
 *    ds_in     - name of segment dataset
 *    first_obs - first row in segment
 *    last_obs  - last row in segment
 *    var       - variable to model
 *    interval  - time interval
 *
 * Output Parameters:
 *    error    - approximation error for segment
 *
 *--------------------------------------------------*/
%macro linear_approximation_macro;

   %let ds_in     = %sysfunc(dequote(&ds_in));
   %let var       = %sysfunc(dequote(&var));
 
   data _temp1_; set &ds_in(firstobs=&first_obs obs=&last_obs);
     retain _TREND_ 0;
     _TREND_ = _TREND_ + 1;
   run;

   proc reg data=_temp1_ outest=_est_ noprint;
      model &var = _TREND_ / sse;
	  output out=_reg_&first_obs._&last_obs. p=PREDICT;
   run; quit;

   proc sql noprint; select _SSE_ into :ERROR from _est_; quit;

%mend linear_approximation_macro;

/*------------------------------------------------*
 * FCMP Function wrappers for above macros
 *------------------------------------------------*/
proc fcmp outlib=sasuser.examples.topdown;

  function get_nobs(ds $);
    nobs = 0;
    rc = run_macro( 'get_nobs_macro', ds, nobs );
    return(nobs);
  endsub;

  subroutine append_segment(seg_ds $, start, end, error);
    first_obs = start;
    last_obs = end;
    rc = run_macro( 'append_segment_macro', 
                    seg_ds, first_obs, last_obs, error );
  endsub;

  function linear_approximation(ds_in $, var $, interval $, first_obs, last_obs);
     error = .;
     rc = run_macro('linear_approximation_macro', 
                    ds_in, first_obs, last_obs, var, interval, error);
     return(error);
  endsub;

run;

/* set cmplib option so our functions can be found */
options cmplib= sasuser.examples;

/*-------------------------------------------------*
 * Top-Down Segmentation Function.  
 *
 * Resursively splits data in half until segments
 * meet threshold criteria. 
 *
 * Parameters:
 *   data      : time-series data to segment 
 *   var       : variable to model
 *   interval  : interval of data
 *   start     : first row (MISSING first time)
 *   end       : last row (MISSING first time)
 *   error     : error for current segment 
 *               (MISSING first time)
 *   threshold : threshold for error improvement 
 *               (percentage)
 *   min_seg   : minimum observations for segment
 *
 * Output Parameters:
 *   seg_data  - dataset to write segment info
 *-------------------------------------------------*/
proc fcmp outlib=sasuser.examples.topdown;

  subroutine segment_topdown(data $, seg_data $, var $, interval $,
                            start_in, end_in, error_in,
                            threshold, min_seg);


    /*- initial setup -*/
    if (start_in eq .) then start = 1; else start=start_in;
    if (end_in eq .) then end = get_nobs(data); else end=end_in;

    /*- calculate error on current segment -*/
    if (error_in eq .) then do;
       error = linear_approximation(data,var,interval,start,end);
    end;
    else do;
       error = error_in;
    end;

	put "segment_topdown called with" start= end= error= ;

    /*- calculate mid point -*/
    mid = start + floor((end-start)/2);

    /*- calculate error on LEFT SEGMENT -*/
    left_error = linear_approximation(data,var,interval,start,mid);

    /*- calculate error on RIGHT SEGMENT -*/
    right_error = linear_approximation(data,var,interval,mid+1,end);

    /*- calculate combined error improvement -*/
    combined_error = left_error + right_error;
    improvement = (error - combined_error) / error;
    splitlength = mid-start;

    put combined_error= improvement=;

    /* recursively split segment */
    if (improvement > threshold AND splitlength/2 >= min_seg) then do;

       /* left segment */
       call segment_topdown(data,seg_data,var,interval,
                           start,mid,left_error,threshold,min_seg);

       /* right segment */
       call segment_topdown(data,seg_data,var,interval,
                           mid+1,end,right_error,threshold,min_seg);

    end;

    /* else store current segment */
    else do;
       call append_segment(seg_data,start,end,error);
    end;

  endsub;

run;

/* make sure there are no old datasets hanging around */
proc datasets library=work memtype=data nolist;
		delete _reg_: segds_: ;
quit;

libname sugi "\\d71235\Public\sugi";
data _NULL_; 
 call segment_topdown("sugi.snp", "work.segds_reg_15", "close", "daily", 
                      ., ., ., 0.15, 10);
 call segment_topdown("sugi.snp", "work.segds_reg_20", "close", "daily", 
                      ., ., ., 0.20, 10);
run;


